(*  Title:  Interpret/refine.thy
    Author: Walther Neuper
   (c) due to copyright terms
*)

theory refine
imports Isac.Build_Isac
begin

text \<open>
  *setup* for a test only; thus it is not contained in Knowledge/*.
  *setup* Know_Store.add_pbls is impossible in *.sml files.
  The respective test/../refine.sml is called in Test_Isac.thy.
\<close> ML \<open>
(*/----- local in struct Problem -------------\*)
val drop_pos_model = (fn (a, b, _) => (a, b))
val drop_pos_where_ = (fn (b, _) => b)
(*\----- local in struct Problem -------------/*)
val thy = @{theory "Test"}
\<close> ML \<open>
(* cp from problem.sml, Outer_Syntax.command \<^command_keyword>\<open>problem\<close> *)
  val arg_1 = Problem.prep_store thy "pbl_test_refine" [] Problem.id_empty
    ([], [], NONE) (["refine", "test"], [], Rule_Set.empty)

  val (model, where_) = Model_Pattern.parse_pos ctxt
    [("#Given", [("fixedValues a_a", Position.none)])]
  val (model', where_') = (map drop_pos_model model, map drop_pos_where_ where_)
  val arg_2 = Problem.prep_store thy "pbl_pbla" [] Problem.id_empty
    (model', where_', NONE) (["pbla", "refine", "test"], [], Rule_Set.empty)

  val (model, where_) = Model_Pattern.parse_pos ctxt
    [("#Given", [("fixedValues a_a", Position.none), ("maximum a_1", Position.none)])]
  val (model', where_') = (map drop_pos_model model, map drop_pos_where_ where_)
  val arg_3 = Problem.prep_store thy "pbl_pbla1" [] Problem.id_empty
    (model', where_', NONE) (["pbla1", "pbla", "refine", "test"], [], Rule_Set.empty)

  val (model, where_) = Model_Pattern.parse_pos ctxt
    [("#Given", [("fixedValues a_a", Position.none), ("valuesFor a_2", Position.none)])]
  val (model', where_') = (map drop_pos_model model, map drop_pos_where_ where_)
  val arg_4 = Problem.prep_store thy "pbl_pbla2" [] Problem.id_empty
    (model', where_', NONE) (["pbla2", "pbla", "refine", "test"], [], Rule_Set.empty)

  val (model, where_) = Model_Pattern.parse_pos ctxt
    [("#Given", [("fixedValues a_a", Position.none),
                 ("valuesFor a_2", Position.none),
                 ("functionOf a2_x", Position.none)])]
  val (model', where_') = (map drop_pos_model model, map drop_pos_where_ where_)
  val arg_5 = Problem.prep_store thy "pbl_pbla2x" [] Problem.id_empty
    (model', where_', NONE) (["pbla2x", "pbla2", "pbla", "refine", "test"], [], Rule_Set.empty)

  val (model, where_) = Model_Pattern.parse_pos ctxt
    [("#Given", [("fixedValues a_a", Position.none),
                 ("valuesFor a_2", Position.none),
                 ("boundVariable a2_y", Position.none)])]
  val (model', where_') = (map drop_pos_model model, map drop_pos_where_ where_)
  val arg_6 = Problem.prep_store thy "pbl_pbla2y" [] Problem.id_empty
    (model', where_', NONE) (["pbla2y", "pbla2", "pbla", "refine", "test"], [], Rule_Set.empty)

  val (model, where_) = Model_Pattern.parse_pos ctxt
    [("#Given", [("fixedValues a_a", Position.none),
                 ("valuesFor a_2", Position.none),
                 ("interval a2_z", Position.none)])]
  val arg_7 = Problem.prep_store thy "pbl_pbla2z" [] Problem.id_empty
    (model', where_', NONE) (["pbla2z", "pbla2", "pbla", "refine", "test"], [], Rule_Set.empty)

  val (model, where_) = Model_Pattern.parse_pos ctxt
    [("#Given", [("fixedValues a_a", Position.none),
                 ("relations a_3", Position.none)])]
  val (model', where_') = (map drop_pos_model model, map drop_pos_where_ where_)
  val arg_8 = Problem.prep_store thy "pbl_pbla3" [] Problem.id_empty
    (model', where_', NONE) (["pbla3", "pbla", "refine", "test"], [], Rule_Set.empty)
\<close>
setup \<open>Know_Store.add_pbls @{context} [arg_1, arg_2, arg_3, arg_4, arg_5, arg_6, arg_7, arg_8]\<close>

(*ML_file "$ISABELLE_ISAC_TEST/Tools/isac/Specify/refine.sml" ... is called in Test_Isac.thy below,
  so do NOT call it HERE; ?circularity*) 

end
